from inspect import isclass
from furthrmind.collection.baseclass import BaseClass
from typing_extensions import TYPE_CHECKING, List, Self

class Category(BaseClass):
    id = ""
    name = ""
    description = ""
    project = ""

    _attr_definition = {
        "project": {"class": "Project"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = Category.fm.get_project_url(project_id)
        url = f"{project_url}/researchcategory/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/researchcategory/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/researchcategory"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/researchcategory"
        return url

    @classmethod
    def get(cls, id=None) -> Self:
        """
        Method to get all one category by it's id
        If called on an instance of the class, the id of the class is used
        :param str id: id of requested category 
        :return Self: Instance of category class
        """

        if isclass(cls):
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        """
        Method to get all categories belonging to one project
        :param str project_id: Optionally to get categories from another project as the furthrmind sdk was initiated with, defaults to None
        :return List[Self]: List with instances of category class
        """
        return super().get_all(project_id)
    
    @staticmethod
    def create():
        pass







import requests
from furthrmind.collection.project import Project
import os
from furthrmind import collection


class Furthrmind:
    Experiment = collection.Experiment
    File = collection.File
    FieldData = collection.FieldData
    ResearchItem = collection.ResearchItem
    Sample = collection.Sample
    Unit = collection.Unit
    ComboBoxEntry = collection.ComboBoxEntry
    Field = collection.Field
    DataTable = collection.DataTable
    Group = collection.Group
    Column = collection.Column
    Project = collection.Project
    Category = collection.Category

    def __init__(self, host, api_key=None, api_key_file=None, project_id=None, project_name=None):

        if not host.startswith("http"):
            host = f"https://{host}"
        self.host = host
        self.base_url = f"{host}/api2"
        self.session = requests.session()

        assert api_key is not None or api_key_file is not None, "Either api_key or api_key_file must be specified"

        if api_key_file:
            assert os.path.isfile(api_key_file), "Api key file is not a valid file"
            with open(api_key_file, "r") as f:
                api_key = f.read()

        self.session.headers.update({"X-API-KEY": api_key})

        self.api_key = api_key
        self.project_id = project_id
        self._write_fm_to_base_class()

        # write project_url with wrong project_id to enable get request if name is provided
        self.project_url = f"{self.base_url}/projects/{self.project_id}"

        if project_name is not None:
            project = Project.get(name=project_name)
            if project:
                self.project_id = project.id

            assert self.project_id is not None, f"Project '{project_name} was not found. Check the spelling"

        # rewrite project_url after project is successully found
        self.project_url = f"{self.base_url}/projects/{self.project_id}"

    def get_project_url(self, project_id=None):
        if project_id is None:
            return self.project_url
        else:
            project_url = self.project_url.replace(str(self.project_id), project_id)
            return project_url

    def get_projects(self):
        return self.get.project()

    def _write_fm_to_base_class(self):
        from furthrmind.collection.baseclass import BaseClass
        BaseClass.fm = self






from furthrmind.collection.baseclass import BaseClass
from typing_extensions import Self, List
from inspect import isclass

class Unit(BaseClass):
    id = ""
    name = ""
    longname = ""
    definition = ""

    _attr_definition = {
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = Unit.fm.get_project_url(project_id)
        url = f"{project_url}/units/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/units/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/units"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/units"
        return url
    
    @classmethod
    def get(cls, id=None) -> Self:
        """
        Method to get all one unit by it's id
        If called on an instance of the class, the id of the class is used
        :param str id: id of requested category 
        :return Self: Instance of unit class
        """
        if isclass(cls):
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        """
        Method to get all units belonging to one project
        :param str project_id: Optionally to get units from another project as the furthrmind sdk was initiated with, defaults to None
        :return List[Self]: List with instances of unit class
        """
        return super().get_all(project_id)

    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls, name: str, definition: str = None, project_id=None) -> Self:
        """
        Method to create a new unit

        :param name: name of the new unit
        :param definition: Unit definition in SI units to convert the new unit to an SI Value. E.g. for unit cm², the
           definition would be: 'cm * cm'. For valid units please check the webapp, open the unit editor.
           You will find there a list of valid units. A definition can als contain scalar values.
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with

        :return: instance of the unit class
        """

        if not name:
            raise ValueError("Name is required")

        data = {"name": name, "definition": definition}
        id = cls.post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator
    def create_many(cls, data_list, project_id=None) -> Self:
        """
        Method to create a new unit

        :param data_list: List of dictionaries with the following keys:
        - name: name of the new unit
        - definition: Unit definition in SI units to convert the new unit to an SI Value. E.g. for unit cm², the
           definition would be: 'cm * cm'. For valid units please check the webapp, open the unit editor.
           You will find there a list of valid units. A definition can als contain scalar values.
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with

        :return: instance of the unit class

        """

        new_data_list = []
        for data in data_list:
            name = data.get("name")
            definition = data.get("definition")
            if not name:
                raise ValueError("Name is required")

            data = {"name": name, "definition": definition}
            new_data_list.append(data)

        id_list = cls.post(new_data_list, project_id)
        for data, id in zip(new_data_list, id_list):
            data["id"] = id

        return new_data_list






from datetime import datetime, date

from bson import ObjectId
from typing_extensions import List, TYPE_CHECKING

from furthrmind.collection.baseclass import BaseClass
from furthrmind.utils import instance_overload

if TYPE_CHECKING:
    from furthrmind.collection.unit import Unit


class FieldData(BaseClass):
    id = ""
    field_name = ""
    field_id = ""
    field_type = ""
    si_value = None
    unit: List["Unit"] = None
    author = None
    value = None

    _attr_definition = {
        "unit": {"class": "Unit"},
        "field_name": {"data_key": "fieldname"},
        "field_type": {"data_key": "fieldtype"},
        "field_id": {"data_key": "fieldid"},
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

        # create instance methods for certain class_methods
        instance_methods = ["_check_value_type"]
        instance_overload(self, instance_methods)

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/fielddata"
        return url

    @classmethod
    def get(cls, id=None):
        raise TypeError("Not implemented")

    @classmethod
    def get_all(cls, project_id=None):
        raise TypeError("Not implemented")

    def update_value(self, value):
        """
        Method to update the value of fielddata

        :param value:
            - Numeric: float or int, or a string convertable to a float
            - Date: datetime, or date object, or unix timestamp or string with iso format
            - SingleLine: string
            - ComboBoxEntry: dict with id or name as key, or string with name, or string with id
            - MultiLine: dict with content as key, or string
            - CheckBox: boolean
        :return: id
        """
        value = self.__class__._check_value_type(value, self.field_type)
        data = {"id": self.id,
                "value": value}
        id = self.post(data)
        self.value = value
        return id

    def set_calculation_result(self, value: dict):
        """
        Method to set a calculation result
        :param value: dict
        :return: id
        """
        if not self.field_type in ["Calculation", "RawDataCalc"]:
            raise TypeError("Only applicable for calculation field")

        url = f"{self.fm.base_url}/set-result/{self.id}"
        response = self.fm.session.post(url, json=value)
        if response.status_code != 200:
            raise ValueError("Setting calculation result failed")
        return self.id

    @classmethod
    def _check_value_type(cls, value, fieldtype=None):
        if value is None:
            return value
        if issubclass(cls, BaseClass):
            # classmethod
            if fieldtype is None:
                raise ValueError("fieldtype must not be None")
        else:
            # instance method
            self = cls
            fieldtype = self.fieldtype
        if fieldtype == "Numeric":
            try:
                value = float(value)
            except:
                raise TypeError("Not numeric")
            return value
        elif fieldtype == "Date":
            if isinstance(value, datetime):
                return int(value.timestamp())
            if isinstance(value, date):
                value = datetime.combine(value, datetime.min.time())
                return int(value.timestamp())
            if isinstance(value, str):
                try:
                    value = datetime.fromisoformat(value)
                    return int(value.timestamp())
                except ValueError:
                    raise TypeError("No iso time format")
            if isinstance(value, int):
                return value
        elif fieldtype == "SingleLine":
            if isinstance(value, str):
                return value
            if isinstance(value, (float, int)):
                return str(value)
            raise TypeError("Type must be string")

        elif fieldtype == "ComboBoxEntry":
            if isinstance(value, dict):
                if "id" in value:
                    return value
                if "name" in value:
                    return value
                raise TypeError("The dict must have either id or name key")
            if isinstance(value, str):
                try:
                    value = ObjectId(value)
                    value = {"id": value}
                except:
                    value = {"name": value}
                return value
            raise TypeError("Only string and dict supported")

        elif fieldtype == "MultiLine":
            if isinstance(value, dict):
                if "content" not in value:
                    raise TypeError("Key 'content' is required")
                return value
            if isinstance(value, str):
                value = {"content": value}
                return value
            raise TypeError("Only string and dict supported")

        elif fieldtype == "CheckBox":
            if not isinstance(value, bool):
                raise TypeError("value must be a bool")
            return value

    def update_unit(self, unit):
        """
        Method to update the unit of fielddata

        :param unit:
            - dict with id or name, or name as string, or id as string
        :return: id
        """
        unit = self._check_unit(unit)
        data = {"id": self.id,
                "unit": unit}
        id = self.post(data)
        self.unit = unit
        return id

    @classmethod
    def _check_unit(cls, unit):
        if not unit:
            return unit
        if isinstance(unit, dict):
            if "id" in unit:
                return unit
            if "name" in unit:
                return unit
            raise TypeError("The dict must have either id or name key")

        elif isinstance(unit, str):
            try:
                unit = ObjectId(unit)
                unit = {"id": unit}
            except:
                unit = {"name": unit}
            return unit
        raise TypeError("Only string and dict supported")

    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls, field_name, field_type, field_id, value, unit, project_id=None):
        """
        Method to create a new fielddata

        :param field_name: name of the field. Either field name and field_type must be specified, or field_id
                           must be specified
        :param field_type: type of the field. Must be out of:
            - Numeric
            - Date
            - SingleLine
            - ComboBoxEntry
            - MultiLine
            - CheckBox
            - Calculation
        :param field_id: id of the field
        :param value:
            - Numeric: float or int, or a string convertable to a float
            - Date: datetime, or date object, or unix timestamp or string with iso format
            - SingleLine: string
            - ComboBoxEntry: dict with id or name as key, or string with name, or string with id
            - MultiLine: dict with content as key, or string
            - CheckBox: boolean
        :param unit: dict with id or name, or name as string, or id as string
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: instance of fielddata class
        """
        from furthrmind.collection import Field

        if field_id:
            data = {"fieldid": field_id}
            field = Field(id=field_id)
            field.get()
            field_type = field.type
        else:
            if not field_name or not field_type:
                raise ValueError("fieldname and fieldtype must be specified")
            data = {"fieldname": field_name,
                    "fieldtype": field_type}

        value = FieldData._check_value_type(value, field_type)
        data["value"] = value

        if unit:
            unit = FieldData._check_unit(unit)
            data["unit"] = unit

        id = FieldData.post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator
    def create_many(cls, data_list, project_id=None):
        """
        Method to create many new fielddata

        :param data_list: List with dicts with the following kneys:
        - field_name: name of the field. Either field name and field_type must be specified, or field_id
            must be specified
        -  field_type: type of the field. Must be out of:
            - Numeric
            - Date
            - SingleLine
            - ComboBoxEntry
            - MultiLine
            - CheckBox
            - Calculation

        - field_id: id of the field
        - value:
            - Numeric: float or int, or a string convertable to a float
            - Date: datetime, or date object, or unix timestamp or string with iso format
            - SingleLine: string
            - ComboBoxEntry: dict with id or name as key, or string with name, or string with id
            - MultiLine: dict with content as key, or string
            - CheckBox: boolean

        - unit: dict with id or name, or name as string, or id as string

        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: list with instances of fielddata class

        """

        from furthrmind.collection import Field

        post_data_list = []
        for data in data_list:
            field_id = data.get("field_id")
            field_name = data.get("field_name")
            field_type = data.get("field_type")
            value = data.get("value")
            unit = data.get("unit")

            if field_id:
                _data = {"fieldid": field_id}
                field = Field(id=field_id)
                field.get()
                field_type = field.type
            else:
                if not field_name or not field_type:
                    raise ValueError("field_name and field_type must be specified")
                _data = {"fieldname": field_name,
                         "fieldtype": field_type}

            value = FieldData._check_value_type(value, field_type)
            _data["value"] = value

            if unit:
                unit = FieldData._check_unit(unit)
                _data["unit"] = unit
            post_data_list.append(_data)

        id_list = FieldData.post(post_data_list, project_id)
        for data, id in zip(data_list, id_list):
            data["id"] = id
        return data_list



from furthrmind.collection.baseclass import (BaseClassWithFieldData, BaseClassWithFiles,
                                             BaseClassWithGroup, BaseClass,
                                             BaseClassWithLinking)
from typing_extensions import Self, Dict, List, TYPE_CHECKING
from furthrmind.utils import instance_overload
from inspect import isclass
if TYPE_CHECKING:
    from furthrmind.collection import *


class Sample(BaseClassWithFieldData,
             BaseClassWithFiles, BaseClassWithGroup,
             BaseClassWithLinking, BaseClass):
    id = ""
    name = ""
    neglect = False
    shortid = ""
    files: List["File"] = []
    fielddata: List["FieldData"] = []
    linked_experiments: List["Experiment"] = []
    linked_samples: List[Self] = []
    linked_researchitems: Dict[str, List["ResearchItem"]] = {}
    groups: List["Group"] = []
    datatables: List["DataTable"] = []

    _attr_definition = {
        "files": {"class": "File"},
        "fielddata": {"class": "FieldData"},
        "groups": {"class": "Group"},
        "linked_samples": {"class": "Sample"},
        "linked_experiments": {"class": "Experiment"},
        "linked_researchitems": {"class": "ResearchItem", "nested_dict": True},
        "datatables": {"class": "DataTable"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)
        # create instance methods for certain class_methods
        instance_methods = ["get"]
        instance_overload(self, instance_methods)

    def _get_url_instance(self, project_id=None):
        project_url = Sample.fm.get_project_url(project_id)
        url = f"{project_url}/samples/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/samples/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/samples"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/samples"
        return url

    @classmethod
    def get(cls, id=None, name=None) -> Self:
        """
        Method to get all one sample by it's id or short_id
        If called on an instance of the class, the id of the class is used
        :param str id: id or short_id of requested sample 
        :param str name: name of requested sample 
        :return Self: Instance of sample class
        """

        if isclass(cls):
            if id is None:
                id = name
            assert id is not None or name is not None, "Either id or name must be specified"
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        """
        Method to get all samples belonging to one project
        :param str project_id: Optionally to get samples from another project as the furthrmind sdk was initiated with, defaults to None
        :return List[Self]: List with instances of sample class
        """
        return super().get_all(project_id)
    
    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls, name, group_name=None, group_id=None, project_id=None) -> Self:
        """
        Method to create a new sample

        :param name: the name of the item to be created
        :param group_name: The name of the group where the new item will belong to. group name can be only considered
            for groups that are not subgroups. Either group_name or group_id must be specified
        :param group_id: the id of the group where the new item will belong to. Either group_name or group_id must be specified
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return instance of the sample class

        """

        return Sample._create(name, group_name, group_id, project_id)

    @classmethod
    @BaseClass._create_instances_decorator
    def create_many(cls, data_list: List[Dict], project_id=None) -> Self:
        """
        Method to create multiple samples

        :param data_list: dict with the following keys:
            - name: the name of the item to be created
            - group_name: The name of the group where the new item will belong to. group name can be only considered
            for groups that are not subgroups. Either group_name or group_id must be specified
            - group_id: the id of the group where the new item will belong to. Either group_name or group_id must be specified
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return list with instance of the sample class

        """

        return Sample._create_many(data_list, project_id)

    def add_datatable(self, name: str, columns: List[Dict]=None, project_id=None ) -> "DataTable":
        """
        Method to create a new datatable within this experiment

        :param name: name of the datatable
        :param columns: a list of columns that should be added to the datatable. List with dicts with the following keys:
            - name: name of the column
            - type: Type of the column, Either "Text" or "Numeric". Data must fit to type, for Text all data
            will be converted to string and for Numeric all data is converted to float (if possible)
            - data: List of column values, must fit to column_type
            - unit: dict with id or name, or name as string, or id as string
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: instance of column datatable class

        """

        from furthrmind.collection import DataTable
        datatable = DataTable.create(name, sample_id=self.id, columns=columns, project_id=project_id)
        return datatable






from ..utils import furthr_wrap
from furthrmind.collection.baseclass import (BaseClassWithFieldData, BaseClassWithFiles,
                                             BaseClassWithGroup, BaseClass,
                                             BaseClassWithLinking)
from typing_extensions import List, Dict, Self, TYPE_CHECKING
from inspect import isclass

if TYPE_CHECKING:
    from furthrmind.collection import *

class ResearchItem(BaseClassWithFieldData, BaseClassWithFiles, BaseClassWithGroup, BaseClassWithLinking, BaseClass ):
    id = ""
    name = ""
    neglect = False
    shortid = ""
    files: List["File"] = []
    fielddata: List["FieldData"] = []
    linked_experiments: List["Experiment"] = []
    linked_samples: List["Sample"] = []
    linked_researchitems: Dict[str, List["ResearchItem"]] = {}
    groups: List["Group"] = []
    category: "Category" = None
    datatables: List["DataTable"] = []

    _attr_definition = {
        "files": {"class": "File"},
        "fielddata": {"class": "FieldData"},
        "groups": {"class": "Group"},
        "linked_samples": {"class": "Sample"},
        "linked_experiments": {"class": "Experiment"},
        "linked_researchitems": {"class": "ResearchItem", "nested_dict": True},
        "datatables": {"class": "DataTable"},
        "category": {"class": "Category"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = ResearchItem.fm.get_project_url(project_id)
        url = f"{project_url}/researchitems/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/researchitems/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/researchitems"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/researchitems"
        return url
    
    @classmethod
    def get(cls, id=None) -> Self:
        """
        Method to get all one researchitem by it's id or short_id
        If called on an instance of the class, the id of the class is used
        :param str id: id or short_id of requested researchitem 
        :return Self: Instance of researchitem class
        """

        if isclass(cls):
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        """
        Method to get all researchitems belonging to one project
        :param str project_id: Optionally to get researchitems from another project as the furthrmind sdk was initiated with, defaults to None
        :return List[Self]: List with instances of researchitem class
        """
        return super().get_all(project_id)

    @classmethod
    def get_by_name(cls, name, category_name, project_id=None) -> Self:
        return cls._get_by_name_class_method(name, category_name, project_id)

    @classmethod
    def _get_by_name_class_method(cls, name, category_name, project_id):
        all_data = cls.get_all(project_id)
        for d in all_data:
            if d.category.name.lower() == category_name.lower():
                if d.name.lower() == name.lower():
                    return d
        
        raise ValueError("No item with this name found")

    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls,name, group_name = None, group_id=None, category_name=None, category_id = None, project_id=None) -> Self:
        """
        Method to create a new researchitem

        :param name: the name of the item to be created
        :param group_name: The name of the group where the new item will belong to. group name can be only considered
            for groups that are not subgroups. Either group_name or group_id must be specified
        :param group_id: the id of the group where the new item will belong to. Either group_name or group_id must be specified
        :category_name: the name of the category that the new item will belong to. Either category_name or category_id must be specified
        :category_id: the id of the category that the new item will belong to. Either category_name or category_id must be specified
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return instance of the researchitem class

        """

        from furthrmind.collection import Category
        if not category_name and not category_id:
            raise ValueError("Either category name or id must be specified")
        data = cls._prepare_data_for_create(name, group_name, group_id, project_id)

        category_dict = {}
        if category_name:
            category_dict["name"] = category_name
        if category_id:
            category_dict["id"] = category_id

        data["category"] = category_dict
        id = cls.post(data, project_id)

        if "id" not in category_dict:
            categories = Category.get_all(project_id)
            for cat in categories:
                if cat.name == category_name:
                    category_dict["id"] = cat.id
                    break

        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator
    def create_many(cls, data_list: List[Dict], project_id=None) -> Self:
        """
        Method to create multiple experiments

        :param data_list: dict with the following keys:
            - name: the name of the item to be created
            - group_name: The name of the group where the new item will belong to. group name can be only considered
            for groups that are not subgroups. Either group_name or group_id must be specified
            - group_id: the id of the group where the new item will belong to. Either group_name or group_id must be specified
            - category_name: the name of the category that the new item will belong to. Either category_name or category_id must be specified
            - category_id: the id of the category that the new item will belong to. Either category_name or category_id must be specified
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return list with instance of the experiment class

        """
        from furthrmind.collection import Group

        new_list = []
        category_id_not_present = False

        groups = Group.get_all()
        for data in data_list:
            category_name = data.get('category_name')
            category_id = data.get('category_id')
            if not category_name and not category_id:
                raise ValueError("Either category name or id must be specified")

            temp_data = cls._prepare_data_for_create(data.get("name"), data.get("group_name"), data.get("group_id"),
                                                     project_id, groups)

            category_dict = {}
            if category_name:
                category_dict["name"] = category_name
            if category_id:
                category_dict["id"] = category_id

            temp_data["category"] = category_dict
            new_list.append(temp_data)
            if not "id" in category_dict:
                category_id_not_present = True

        id_list = cls.post(new_list, project_id)
        category_mapping = {}
        if category_id_not_present:
            categories = Category.get_all(project_id)
            category_mapping = {cat.name: cat for cat in categories}

        for data, id in zip(new_list, id_list):
            data["id"] = id
            if "id" not in data["category"]:
                cat_id = category_mapping.get(data["category"]["name"])
                data["category"]["id"] = cat_id

        return new_list

    def add_datatable(self, name: str, columns: List[Dict], project_id=None ) -> "DataTable":
        """
        Method to create a new datatable within this experiment

        :param name: name of the datatable
        :param columns: a list of columns that should be added to the datatable. List with dicts with the following keys:
            - name: name of the column
            - type: Type of the column, Either "Text" or "Numeric". Data must fit to type, for Text all data
            will be converted to string and for Numeric all data is converted to float (if possible)
            - data: List of column values, must fit to column_type
            - unit: dict with id or name, or name as string, or id as string
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: instance of column datatable class

        """

        from furthrmind.collection import DataTable
        datatable = DataTable.create(name, researchitem_id=self.id, columns=columns, project_id=project_id)
        return datatable




def get_collection_class(collection_name):
    return eval(collection_name)

from .file import File
from .fielddata import FieldData
from .researchitem import ResearchItem
from .sample import Sample
from .unit import Unit
from .comboboxentry import ComboBoxEntry
from .field import Field
from .datatable import DataTable
from .group import Group
from .experiment import Experiment
from .column import Column
from .project import Project
from .category import Category






from furthrmind.collection.baseclass import BaseClass
from typing_extensions import List, Dict, Self, TYPE_CHECKING
from inspect import isclass
if TYPE_CHECKING:
    from furthrmind.collection import *

class Project(BaseClass):
    id = ""
    name = ""
    info = ""
    shortid = ""
    samples: List["Sample"] = []
    experiments: List["Experiment"] = []
    groups: List["Group"] = []
    units: List["Unit"] = []
    researchitems: Dict[str, List["ResearchItem"]] = {}
    permissions = []
    fields: List["Field"] = []

    _attr_definition = {
        "samples": {"class": "Sample"},
        "experiments": {"class": "Experiment"},
        "groups": {"class": "Group"},
        "units": {"class": "Unit"},
        "researchitems": {"class": "ResearchItem", "nested_dict": True},
        "fields": {"class": "Field"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self):
        project_url = Column.fm.get_project_url(self.id)
        return project_url

    @classmethod
    def _get_url_class(cls, id):
        project_url = cls.fm.get_project_url(id)
        return project_url

    @classmethod
    def _get_all_url(cls):
        return f"{cls.fm.base_url}/projects"

    @classmethod
    def _post_url(cls):
        return f"{cls.fm.base_url}/projects"
    
    @classmethod
    def get(cls, id=None, name=None) -> Self:
        """
        Method to get all one project by it's id
        If called on an instance of the class, the id of the class is used
        :param str id: id or short_id of requested project 
        :param str name: name of requested project 
        :return Self: Instance of project class
        """

        if isclass(cls):
            if id is None:
                id = name
            assert id is not None or name is not None, "Either id or name must be specified"
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls) -> List[Self]:
        """
        Method to get all projects belonging to one project
        :return List[Self]: List with instances of project class
        """
        return super().get_all()

    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls, name: str) -> Self:
        """
        Method to create a new project
        :param name: Name of the new project
        :return: instance of the project class
        """

        if not name:
            raise ValueError("Name is required")
        data = {"name": name}
        id = cls.post(data)
        data["id"] = id
        return data




